import re
import redis
import fakeredis
from redorm.settings import REDORM_URL
from redorm.exceptions import DevelopmentServerUsedInProd


class RedormClient:
    pool: redis.ConnectionPool
    client: redis.Redis

    def __init__(self):
        self.server = None
        self.bind(REDORM_URL)

    def bind(self, url) -> bool:
        match = re.match(
            r"redis://(?P<host>[^:^/]+)(?::(?P<port>[0-9]+))?(?:/(?P<db>[0-9]+))?", url,
        )
        if match:
            d = match.groupdict()
            host = d["host"]
            port = int(d["port"]) if d["port"] is not None else 6379
            db = int(d["db"]) if d["db"] is not None else 0
            self.pool = redis.ConnectionPool(host="localhost", port=6379, db=1)
            self.client = redis.Redis(connection_pool=self.pool)
            return True
        else:
            # Lightweight fake redis
            self.server = fakeredis.FakeServer()
            self.client = fakeredis.FakeStrictRedis(server=self.server)
            return False

    def init_app(self, app):
        url = app.config.get("REDORM_URL")
        if url:
            success = self.bind(url)
            if (
                not success
                and app.config.get("FLASK_ENV") == "production"
                and not app.config.get("REDORM_INBUILT")
            ):
                raise DevelopmentServerUsedInProd(
                    "Redorm was not provided a Redis URL and would fall back to the in-built Redis. "
                    "In the flask app config either set REDORM_URL to a valid redis connection string "
                    "or set REDORM_INBUILT to True (not recommended)"
                )


red = RedormClient()



from typing import List, Type, Optional, TypeVar, Union
from enum import Enum, auto
from redorm.model import RedisBase, all_models
from redorm.client import red

__all__ = [
    "RelationshipConfigEnum",
    "Relationship",
    "one_to_many",
    "one_to_one",
    "many_to_one",
    "many_to_many",
]


class RelationshipConfigEnum(Enum):
    ONE_TO_MANY = auto()
    ONE_TO_ONE = auto()
    MANY_TO_ONE = auto()
    MANY_TO_MANY = auto()


T = TypeVar("T", bound=RedisBase)
U = TypeVar("U", bound=RedisBase)


class Relationship:
    def __init__(
        self,
        foreign_type: Union[str, Type[U]],
        config: RelationshipConfigEnum = RelationshipConfigEnum.MANY_TO_MANY,
        backref: Optional[str] = None,
    ):
        self.config = config
        if isinstance(foreign_type, str):
            self.foreign_type = None
            self.__foreign_type = foreign_type
        else:
            self.foreign_type = foreign_type
            self.__foreign_type = foreign_type.__name__
        self.fdel = None
        self.backref = backref
        self.__relationship_name = None
        self.__doc__ = None

    def ensure_relationship_name(self, instance: T):
        if self.__relationship_name is None:
            rels = [k for k, v in instance.__class__.__dict__.items() if v is self]
            if len(rels) == 1:
                self.__relationship_name = rels[0]
            elif len(rels) > 1:
                raise ValueError("Same relationship re-used on object")
            else:
                raise ValueError("Could not find relationship on object")

    def get_foreign_type(self) -> Type[U]:
        if self.foreign_type is None:
            self.foreign_type = all_models[self.__foreign_type]
            return self.foreign_type
        else:
            return self.foreign_type

    def fget(self, instance: T):
        self.ensure_relationship_name(instance)
        foreign_type: Type[U] = self.get_foreign_type()
        if self.config in {
            RelationshipConfigEnum.MANY_TO_MANY,
            RelationshipConfigEnum.ONE_TO_MANY,
        }:
            related_ids = red.client.smembers(
                f"{instance.__class__.__name__}:relationship:{self.__relationship_name}:{instance.id}"
            )
            return foreign_type.get_bulk([rid.decode() for rid in related_ids])
        else:
            related_id = red.client.get(
                f"{instance.__class__.__name__}:relationship:{self.__relationship_name}:{instance.id}"
            )
            return (
                foreign_type.get(related_id.decode())
                if related_id is not None
                else None
            )

    def fset(
        self, instance: T, value: Union[None, str, U, List[Union[str, U]]],
    ):
        self.ensure_relationship_name(instance)
        foreign_type = self.get_foreign_type()
        if self.config in {
            RelationshipConfigEnum.MANY_TO_ONE,
            RelationshipConfigEnum.ONE_TO_ONE,
        }:
            related_id_new: Optional[str]
            if isinstance(value, RedisBase):
                related_id_new = value.id
            elif isinstance(value, str):
                related_id_new = value
            elif value is None:
                related_id_new = None
            else:
                raise ValueError("Expected new value of string or Model")
            if related_id_new is None:
                pipeline = red.client.pipeline()
                pipeline.get(
                    f"{instance.__class__.__name__}:relationship:{self.__relationship_name}:{instance.id}"
                )
                pipeline.delete(
                    f"{instance.__class__.__name__}:relationship:{self.__relationship_name}:{instance.id}"
                )
                result = pipeline.execute()
                print(f"results: {result!r}")
                related_id_old = result[0].decode() if result[0] else None
            else:
                related_id_old = red.client.getset(
                    f"{instance.__class__.__name__}:relationship:{self.__relationship_name}:{instance.id}",
                    related_id_new,
                )
            if related_id_new == related_id_old or self.backref is None:
                return
            if self.config == RelationshipConfigEnum.MANY_TO_ONE:
                if related_id_old is None:
                    red.client.sadd(
                        f"{foreign_type.__name__}:relationship:{self.backref}:{related_id_new}",
                        instance.id,
                    )
                elif related_id_new is None:
                    red.client.srem(
                        f"{foreign_type.__name__}:relationship:{self.backref}:{related_id_old}",
                        instance.id,
                    )
                else:
                    red.client.smove(
                        f"{foreign_type.__name__}:relationship:{self.backref}:{related_id_old}",
                        f"{foreign_type.__name__}:relationship:{self.backref}:{related_id_new}",
                        instance.id,
                    )
            else:
                if related_id_old is None:
                    red.client.set(
                        f"{foreign_type.__name__}:relationship:{self.backref}:{related_id_new}",
                        instance.id,
                    )
                elif related_id_new is None:
                    red.client.delete(
                        f"{foreign_type.__name__}:relationship:{self.backref}:{related_id_new}",
                    )
                else:
                    red.client.rename(
                        f"{foreign_type.__name__}:relationship:{self.backref}:{related_id_old}",
                        f"{foreign_type.__name__}:relationship:{self.backref}:{related_id_new}",
                    )
        else:
            if (not isinstance(value, list)) and (not isinstance(value, set)):
                raise ValueError("Expected list or set for new relationships")
            old_related_ids = {
                r.decode()
                for r in red.client.smembers(
                    f"{instance.__class__.__name__}:relationship:{self.__relationship_name}:{instance.id}"
                )
            }
            new_related_ids = {(r.id if isinstance(r, RedisBase) else r) for r in value}
            if len(old_related_ids.symmetric_difference(new_related_ids)) == 0:
                return
            pipeline = red.client.pipeline()
            pipeline.delete(
                f"{instance.__class__.__name__}:relationship:{self.__relationship_name}:{instance.id}"
            )
            if new_related_ids:
                pipeline.sadd(
                    f"{instance.__class__.__name__}:relationship:{self.__relationship_name}:{instance.id}",
                    *new_related_ids,
                )
            if self.backref is None:
                pipeline.execute()
                return
            ids_to_remove = old_related_ids - new_related_ids
            ids_to_add = new_related_ids - old_related_ids
            if self.config == RelationshipConfigEnum.MANY_TO_MANY:
                for idr in ids_to_remove:
                    pipeline.srem(
                        f"{foreign_type.__name__}:relationship:{self.backref}:{idr}",
                        instance.id,
                    )
                for ida in ids_to_add:
                    pipeline.sadd(
                        f"{foreign_type.__name__}:relationship:{self.backref}:{ida}",
                        instance.id,
                    )
            elif self.config == RelationshipConfigEnum.ONE_TO_MANY:
                for idr in ids_to_remove:
                    pipeline.delete(
                        f"{foreign_type.__name__}:relationship:{self.backref}:{idr}"
                    )
                for ida in ids_to_add:
                    pipeline.set(
                        f"{foreign_type.__name__}:relationship:{self.backref}:{ida}",
                        instance.id,
                    )
            else:
                raise ValueError(
                    "Expected relationship config to be of type RelationshipConfigEnum"
                )
            pipeline.execute()

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        if self.fget is None:
            raise AttributeError("unreadable attribute")
        return self.fget(obj)

    def __set__(self, obj, value):
        if self.fset is None:
            raise AttributeError("can't set attribute")
        self.fset(obj, value)

    def __delete__(self, obj):
        if self.fdel is None:
            raise AttributeError("can't delete relationship")
        self.fdel(obj)

    def getter(self, fget):
        return type(self)(fget, self.fset, self.fdel, self.__doc__)

    def setter(self, fset):
        return type(self)(self.fget, fset, self.fdel, self.__doc__)

    def deleter(self, fdel):
        return type(self)(self.fget, self.fset, fdel, self.__doc__)


def one_to_many(
    foreign_type: Union[str, Type[U]], backref: Optional[str] = None,
):
    return Relationship(
        foreign_type=foreign_type,
        config=RelationshipConfigEnum.ONE_TO_MANY,
        backref=backref,
    )


def one_to_one(
    foreign_type: Union[str, Type[U]], backref: Optional[str] = None,
):
    return Relationship(
        foreign_type=foreign_type,
        config=RelationshipConfigEnum.ONE_TO_ONE,
        backref=backref,
    )


def many_to_one(
    foreign_type: Union[str, Type[U]], backref: Optional[str] = None,
):
    return Relationship(
        foreign_type=foreign_type,
        config=RelationshipConfigEnum.MANY_TO_ONE,
        backref=backref,
    )


def many_to_many(
    foreign_type: Union[str, Type[U]], backref: Optional[str] = None,
):
    return Relationship(
        foreign_type=foreign_type,
        config=RelationshipConfigEnum.MANY_TO_MANY,
        backref=backref,
    )



from typing import List, Type, Optional, TypeVar
from dataclasses import dataclass, field
from uuid import uuid4
import json
from dataclasses_jsonschema import JsonSchemaMixin
from redorm.exceptions import RedisInstanceNotFound
from redorm.client import red


S = TypeVar("S", bound="RedisBase")

all_models = dict()


@dataclass
class RedisBase(JsonSchemaMixin):
    id: str = field()

    @classmethod
    def get(cls: Type[S], instance_id: str) -> Optional[S]:
        data = red.client.get(f"{cls.__name__}:member:{instance_id}")
        if data is None:
            return None
        else:
            return cls.from_dict(json.loads(data.decode("utf-8")))

    @classmethod
    def get_bulk(cls: Type[S], instance_ids: List[str]) -> List[S]:
        data = red.client.mget(
            [f"{cls.__name__}:member:{instance_id}" for instance_id in instance_ids]
        )
        if data is None:
            return []
        else:
            return [
                cls.from_dict(json.loads(d.decode("utf-8")))
                for d in data
                if d is not None
            ]

    @classmethod
    def create(cls: Type[S], **kwargs) -> S:
        non_relations = {k: v for k, v in kwargs.items() if k in cls.__annotations__}
        new_instance = cls(id=str(uuid4()), **non_relations)
        for k, v in kwargs.items():
            if k not in cls.__annotations__:
                setattr(new_instance, k, v)

        pipe = red.client.pipeline()
        instance_dict = new_instance.to_dict()
        pipe.set(
            f"{cls.__name__}:member:{new_instance.id}",
            json.dumps(instance_dict, sort_keys=True),
        )
        pipe.execute()
        return new_instance

    @classmethod
    def list(cls: Type[S]) -> List[S]:
        members_data = red.client.mget(
            list(red.client.scan_iter(f"{cls.__name__}:member:*"))
        )
        return [
            cls.from_json(member_data.decode("utf-8"))
            for member_data in members_data
            if member_data
        ]

    def delete(self):
        with red.client.lock(f"{self.__class__.__name__}:member:{self.id}"):
            red.client.delete(self.id)

    def refresh(self) -> None:
        latest = red.client.get(f"{self.__class__.__name__}:member:{self.id}")
        if latest is None:
            raise RedisInstanceNotFound
        for k, v in json.loads(latest.decode("utf-8")):
            setattr(self, k, v)

    def save(self) -> None:
        red.client.set(
            f"{self.__class__.__name__}:member:{self.id}",
            json.dumps(self.to_dict(), sort_keys=True),
        )

    def update(self, **kwargs):
        with red.client.lock(f"{self.__class__.__name__}:member:{self.id}"):
            self.refresh()
            for k, v in kwargs:
                setattr(self, k, v)
            red.client.set(
                f"{self.__class__.__name__}:member:{self.id}",
                json.dumps(self.to_dict(), sort_keys=True),
            )

    def __init_subclass__(cls, **kwargs):
        all_models[cls.__name__] = cls
        super().__init_subclass__(**kwargs)



from redorm.model import all_models, RedisBase, red
from redorm.relationships import (
    many_to_many,
    many_to_one,
    one_to_one,
    one_to_many,
    Relationship,
    RelationshipConfigEnum,
)
from redorm.exceptions import RedisInstanceNotFound, RedormException



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fr:
    requirements = [
        l.strip() for l in fr.read().split("\n") if len(l) > 0 and "#" not in l
    ]

setuptools.setup(
    name="redorm",
    version="0.1.0",
    author="Jack Adamson",
    author_email="jack@mrfluffybunny.com",
    description="A simple redis ORM",
    install_requires=requirements,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jackadamson/redorm",
    packages=setuptools.find_packages(exclude=["tests"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires=">=3.6",
)
